#------------------------------------------------------------------------------
# Name:         test_domain.py
# Purpose:      Test the Domain class
#
# Author:       Anton Korosov
#
# Created:      29.09.2014
# Copyright:    (c) NERSC
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
#------------------------------------------------------------------------------
import os
import sys
import warnings
import unittest

import numpy as np
from mock import patch, PropertyMock, Mock, MagicMock, DEFAULT

from nansat.nsr import NSR
from nansat.vrt import VRT
from nansat.domain import Domain
from nansat.utils import gdal, ogr
from nansat.tests import nansat_test_data as ntd
from nansat.exceptions import NansatProjectionError
from nansat.warnings import NansatFutureWarning

warnings.simplefilter("always", NansatFutureWarning)
warnings.simplefilter("always", UserWarning)

EXTENT_TE_TS = "-te 25 70 35 72 -ts 500 500"
EXTENT_DICT_TE_TS = {'te': [25.0, 70.0, 35.0, 72.0], 'ts': [500.0, 500.0]}
EXTENT_LLE_TS = "-lle 25 70 35 72 -ts 500 500"
EXTENT_DICT_LLE_TS = {'lle': [25.0, 70.0, 35.0, 72.0], 'ts': [500.0, 500.0]}
EXTENT_DICT_LLE_TE_TS = {'lle': [25.0, 70.0, 35.0, 72.0],
                         'te': [25.0, 70.0, 35.0, 72.0], 'ts': [500.0, 500.0]}
GEO_TRANSFORM = [25.0, 0.02, 0.0, 72.0, 0.0, -0.004]
RASTER_X_SIZE = 500
RASTER_Y_SIZE = 500
SRS_PROJ4 = "+proj=latlong +datum=WGS84 +ellps=WGS84 +no_defs"
SRS_EPSG = 4326
EXTENT_BERGEN = "-te 5 60 6 61 -ts 500 500"
EXTENT_WESTCOAST = "-te 1 58 6 64 -ts 500 500"
EXTENT_NORWAY = "-te 3 55 30 72 -ts 500 500"
EXTENT_PARIS = "-te 2 48 3 49 -ts 500 500"


class DomainTest(unittest.TestCase):

    def setUp(self):
        self.test_file = os.path.join(ntd.test_data_path, 'gcps.tif')
        self.test_file_projected = os.path.join(ntd.test_data_path, 'stere.tif')
        if not os.path.exists(self.test_file) or not os.path.exists(self.test_file_projected):
            raise ValueError('No test data available')

    def test_dont_init_from_invalid_combination(self):
        self.assertRaises(ValueError, Domain)
        self.assertRaises(ValueError, Domain, None)
        with self.assertRaises(ValueError):
            Domain(ds=gdal.Open(self.test_file),
                   srs="+proj=latlong +datum=WGS84 +ellps=WGS84 +no_defs",
                   ext="-te 25 70 35 72 -ts 500 500")

    def test_init_from_GDALDataset(self):
        d = Domain(ds=gdal.Open(self.test_file))
        self.assertEqual(type(d), Domain)

    def test_init_from_GDALDataset_and_srs(self):
        d = Domain(ds=gdal.Open(self.test_file),
                   srs="+proj=latlong +datum=WGS84 +ellps=WGS84 +no_defs")
        self.assertEqual(type(d), Domain)

    @patch('nansat.domain.gdal')
    def test_dont_init_if_gdal_AutoCreateWarpedVRT_fails(self, mock_gdal):
        mock_gdal.AutoCreateWarpedVRT.return_value = None
        with self.assertRaises(NansatProjectionError):
            Domain(ds=gdal.Open(self.test_file),
                   srs="+proj=latlong +datum=WGS84 +ellps=WGS84 +no_defs")

    @patch.object(Domain, '_create_extent_dict',
        return_value={'te': [25.0, 70.0, 35.0, 72.0], 'ts': [500.0, 500.0]})
    @patch.object(Domain, '_get_geotransform',
        return_value=([25.0, 0.02, 0.0, 72.0, 0.0, -0.004], 500, 500))
    def test_init_from_srs_and_ext_te(self, mock__get_geotransform, mock__create_extent_dict):
        d = Domain(srs=4326,
                   ext="-te 25 70 35 72 -ts 500 500")
        self.assertEqual(type(d), Domain)

    @patch.object(Domain, '_create_extent_dict',
        return_value={'lle': [25.0, 70.0, 35.0, 72.0], 'ts': [500.0, 500.0]})
    @patch.object(Domain, '_convert_extentDic',
        return_value={'lle': [25.0, 70.0, 35.0, 72.0], 'te': [25.0, 70.0, 35.0, 72.0], 'ts': [500.0, 500.0]})
    @patch.object(Domain, '_get_geotransform',
        return_value=([25.0, 0.02, 0.0, 72.0, 0.0, -0.004], 500, 500))
    def test_init_from_srs_and_ext_lle(self, mock__get_geotransform, mock__convert_extentDic,
                                       mock__create_extent_dict):
        d = Domain(srs=4326,
                   ext="-lle 25 70 35 72 -ts 500 500")
        self.assertEqual(type(d), Domain)

    def test_init_lonlat(self):
        lat, lon = np.mgrid[-90:90:0.5, -180:180:0.5]
        with warnings.catch_warnings(record=True) as recorded_warnings:
            d = Domain(lon=lon, lat=lat)

        nansat_warning_raised = False
        for rw in recorded_warnings:
            if rw.category == NansatFutureWarning:
                nansat_warning_raised = True
        self.assertTrue(nansat_warning_raised)

        self.assertEqual(type(d), Domain)
        self.assertEqual(d.shape(), lat.shape)

    def test_init_from_lonlat(self):
        lat, lon = np.mgrid[-10:10:0.5, -20:20:2]
        d = Domain.from_lonlat(lon=lon, lat=lat)
        self.assertEqual(type(d), Domain)
        self.assertEqual(d.shape(), lat.shape)
        self.assertEqual(len(d.vrt.dataset.GetGCPs()), 100)

    def test_init_from_lonlat_no_gcps(self):
        lat, lon = np.mgrid[-10:10:0.5, -20:20:2]
        d = Domain.from_lonlat(lon=lon, lat=lat, add_gcps=False)
        self.assertEqual(type(d), Domain)
        self.assertEqual(d.shape(), lat.shape)
        self.assertEqual(len(d.vrt.dataset.GetGCPs()), 0)

    @patch.object(Domain, 'get_corners',
        return_value=(np.array([ 25.,  25.,  35.,  35.]), np.array([ 72.,  70.,  72.,  70.])))
    def test_repr(self, mock_get_corners):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        result = d.__repr__()
        test = ('Domain:[500 x 500]\n'
                '----------------------------------------\n'
                'Projection(dataset):\nGEOGCS["WGS 84",\n'
                '    DATUM["WGS_1984",\n'
                '        SPHEROID["WGS 84",6378137,298.257223563]],\n'
                '    PRIMEM["Greenwich",0],\n'
                '    UNIT["degree",0.0174532925199433]]\n'
                '----------------------------------------\n'
                'Corners (lon, lat):\n'
                '\t ( 25.00,  72.00)  ( 35.00,  72.00)\n'
                '\t ( 25.00,  70.00)  ( 35.00,  70.00)\n' )
        self.assertIsInstance(result, str)
        self.assertEquals(result, test)

    def test_write_kml(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        tmpfilename = os.path.join(ntd.tmp_data_path, 'domain_write_kml.kml')
        d.write_kml(kmlFileName=tmpfilename)
        self.assertTrue(os.path.exists(tmpfilename))

    #def test__get_border_kml(self):


    #def test_write_kml_image(self):


    @patch.object(Domain, 'transform_points')
    def test_get_geolocation_grids_from_GDAL_transformer(self, mock_transform_points):
        mock_transform_points.return_value=(
                      np.meshgrid(range(0,500),range(0,500))[0].flatten()*(35-25)/500.+25,
                      np.meshgrid(range(0,500),range(0,500))[1].flatten()*(70-72)/500.+72)
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        lon, lat = d.get_geolocation_grids()
        self.assertEqual(type(lon), np.ndarray)
        self.assertEqual(type(lat), np.ndarray)
        self.assertEqual(lat.shape, (500, 500))

    def test_get_geolocation_grids_from_geolocationArray(self):
        lat, lon = np.mgrid[25:35:0.02, 70:72:0.004]
        d = Domain(lon=lon, lat=lat)
        lon, lat = d.get_geolocation_grids()
        self.assertEqual(type(lon), np.ndarray)
        self.assertEqual(type(lat), np.ndarray)
        self.assertEqual(lat.shape, (500, 500))




    def test_convert_extentDic(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        result = d._convert_extentDic(NSR(4326), {'lle': [25.0, 70.0, 35.0, 72.0], 'ts': [500.0, 500.0]})
        self.assertEqual(result, {'lle': [25.0, 70.0, 35.0, 72.0], 'te': [25.0, 70.0, 35.0, 72.0], 'ts': [500.0, 500.0]})

    def test_add_to_dict(self):
        input_1 = ['-te', '5', '60', '6', '61.1']
        output_1 = {'te': [5., 60., 6., 61.1]}
        key_1, extent_1 = Domain._add_to_dict(dict(), input_1)
        self.assertIsInstance(extent_1, dict)
        self.assertIsInstance(key_1, str)
        self.assertEqual(key_1, input_1[0].replace('-', ''))
        self.assertEqual(len(extent_1), 1)
        self.assertIsInstance(list(extent_1.values()), list)

        for el in list(extent_1.values())[0]:
            self.assertIsInstance(el, float)

        self.assertEqual(extent_1, output_1)
        input_2 = ['-te', '5', 'str', '6', '61']
        with self.assertRaises(ValueError) as opt_err:
            key_2, extent_2 = Domain._add_to_dict(dict(), input_2)
            self.assertEqual(opt_err.args[0], 'Input values must be int or float')

    def test_validate_ts_tr(self):
        input_1 = [100, 200]
        input_2 = ([0, 0], [0, 50], [10, 0], [-1, 10], [10, -100], [-100, -100])
        input_3 = [10]
        self.assertEqual(Domain._validate_ts_tr(input_1), None)
        for inp in input_2:
            with self.assertRaises(ValueError) as opt_err:
                Domain._validate_ts_tr(inp)
                self.assertEqual(opt_err.args[0], 'Resolution or width and height must be bigger '
                                                  'than 0: <-tr x_resolution y_resolution> or '
                                                  '<-ts width height>')
        with self.assertRaises(ValueError) as opt_err:
            Domain._validate_ts_tr(input_3)
            self.assertEqual(opt_err.args[0], '-ts and -tr requires exactly 2 parameters '
                                              '(1 given): <-tr x_resolution y_resolution> or '
                                              '<-ts width height>')
    def test_validate_te_lle(self):
        input_1 = [5., 60., 6., 61.]
        input_2 = ([5., 60., 5., 61.], [5., 60., 4., 61.], [5., 60., 6., 59.])
        input_3 = [60., 5., 61.]
        self.assertEqual(Domain._validate_te_lle(input_1), None)
        for inp in input_2:
            with self.assertRaises(ValueError) as opt_err:
                Domain._validate_te_lle(inp)
                self.assertEqual(opt_err.message, 'Min cannot be bigger than max: '
                                                  '<-te x_min y_min x_max y_max> or '
                                                  '<-lle min_lon min_lat max_lon max_lat>')
        with self.assertRaises(ValueError) as opt_err:
            Domain._validate_te_lle(input_3)
            self.assertEqual(opt_err.args[0], '-te and -lle requires exactly 4 parameters '
                                              '(3 given): <-te x_min y_min x_max y_max> or <-lle'
                                              ' min_lon min_lat max_lon max_lat>')

    def test_check_size(self):
        te_lle_example = '<-te x_min y_min x_max y_max> or <-lle min_lon min_lat max_lon max_lat>'
        tr_ts_example = '<-tr x_resolution y_resolution> or <-ts width height>'
        self.assertEqual(Domain._check_size(2, 2, ('-te', '-lle'), te_lle_example), None)
        with self.assertRaises(ValueError) as opt_err:
            Domain._check_size(1, 2, ('-ts', '-tr'), tr_ts_example)
            self.assertEqual(opt_err.args[0], '-ts and -tr requires exactly 2 parameters '
                                              '(1 given): <-tr x_resolution y_resolution> or '
                                              '<-ts width height>')
        with self.assertRaises(ValueError) as opt_err:
            Domain._check_size(2, 4, ('-te', '-lle'), te_lle_example)
            self.assertEqual(opt_err.args[0], '-te and -lle requires exactly 4 parameters '
                                              '(2 given): <-te x_min y_min x_max y_max> or <-lle'
                                              ' min_lon min_lat max_lon max_lat>')

    def test_gen_regexp(self):
        test_1 = '(-te|-lle)(\\s+[-+]?\\d*[.\\d*]*)(\\s+[-+]?\\d*[.\\d*]*)(\\s+[-+]?' \
                 '\\d*[.\\d*]*)(\\s+[-+]?\\d*[.\\d*]*)\\s?'
        result_1 = Domain._gen_regexp('te', 'lle', 4)
        self.assertIsInstance(result_1, str)
        self.assertEqual(result_1, test_1)
        test_2 = '(-ts|-tr)(\\s+[-+]?\\d*[.\\d*]*)(\\s+[-+]?\\d*[.\\d*]*)\\s?'
        result_2 = Domain._gen_regexp('ts', 'tr', 2)
        self.assertEqual(result_2, test_2)

    def test_create_extent_dict(self):
        test = ('-te 5 60 6 61.1 -ts 500 500',
                '-te -92.08 26.85 -92.00 26.91 -ts 200 200',
                '-te 5 60 6 61.1',
                '-te 5 60 6 61.1 -te 5 60 6 61.1')
        output_1 = {'te': [5., 60., 6., 61.1], 'ts': [500, 500]}
        output_2 = {'te': [-92.08, 26.85, -92.00, 26.91], 'ts': [200, 200]}
        result_1 = Domain._create_extent_dict(test[0])
        self.assertIsInstance(result_1, dict)
        self.assertEqual(len(list(result_1.keys())), 2)
        self.assertEqual(result_1, output_1)
        result_2 = Domain._create_extent_dict(test[1])
        self.assertEquals(result_2, output_2)
        with self.assertRaises(ValueError) as opt_err:
            Domain._create_extent_dict(test[2])
            self.assertEquals(opt_err.args[0], '<extent_dict> must contains exactly 2 parameters ')
        self.assertEqual(result_2, output_2)

        try:
            test = Domain._create_extent_dict(test[2])
        except ValueError as opt_err:
            self.assertEqual(opt_err.args[0], '<extent_dict> must contains exactly 2 parameters '
                                               '("-te" or "-lle") and ("-ts" or "-tr")')
        with self.assertRaises(ValueError) as opt_err:
            Domain._create_extent_dict(test[3])
            self.assertEquals(opt_err.args[0], '<extent_dict> must contains exactly 2 parameters '
                                               '("-te" or "-lle") and ("-ts" or "-tr")')

    def test_get_border(self):
        dom = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        result = dom.get_border(n_points=10)
        lat, lon = result
        self.assertEqual(type(lat), np.ndarray)
        self.assertEqual(type(lon), np.ndarray)
        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 2)
        test_x = [25., 26., 27., 28., 29., 30., 31., 32., 33., 34., 35.,
                  35., 35., 35., 35., 35., 35., 35., 35., 35., 35., 35.,
                  35., 34., 33., 32., 31., 30., 29., 28., 27., 26., 25.,
                  25., 25., 25., 25., 25., 25., 25., 25., 25., 25., 25.]
        test_y = [72., 72., 72., 72., 72., 72., 72., 72.,  72.,
                   72. ,  72. ,  72. ,  71.8,  71.6,  71.4,  71.2,  71. ,  70.8,
                   70.6,  70.4,  70.2,  70. ,  70. ,  70. ,  70. ,  70. ,  70. ,
                   70. ,  70. ,  70. ,  70. ,  70. ,  70. ,  70. ,  70.2,  70.4,
                   70.6,  70.8,  71. ,  71.2,  71.4,  71.6,  71.8,  72.]
        self.assertEqual(list(lat), test_x)
        self.assertEqual(list(lon), test_y)

    def test_compound_row_col_vectors(self):
        result = Domain._compound_row_col_vectors(30, 40, list(range(0, 33, 3)), list(range(0, 44, 4)))
        output_col, output_row = result
        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 2)
        test_col = [0,  3,  6,  9,  12,  15,  18,  21,  24,  27,  30,  30,  30,  30,  30,  30,
                    30,  30,  30,  30,  30,  30,  30,  27,  24,  21,  18,  15,  12,  9,  6,  3,
                    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0]
        test_row = [0,  0,  0,  0,  0,  0,  0, 0,  0,  0,  0,  0,  4,  8,  12,  16,  20,  24,
                    28,  32,  36,  40,  40,  40,  40,  40,  40,  40,  40,  40,  40,  40,  40,
                    40,  36,  32,  28,  24,  20,  16,  12,  8,  4,  0]

        self.assertEqual(output_col, test_col)
        self.assertEqual(output_row, test_row)

    def test_get_row_col_vector(self):
        test_1 = Domain._get_row_col_vector(250, 500)
        self.assertIsInstance(test_1, list)
        self.assertEqual(test_1, list(range(251)))
        self.assertEqual(len(test_1), 251)
        test_2 = Domain._get_row_col_vector(500, 10)
        self.assertEqual(test_2, list(range(0, 550, 50)))
        self.assertEqual(len(test_2), 10 + 1)

    def test_get_border_wkt(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        bwkt = d.get_border_wkt()
        self.assertEqual(type(bwkt), str)
        self.assertTrue('POLYGON' in bwkt)

    def test_get_border_geometry(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        geom = d.get_border_geometry()
        self.assertEqual(type(geom), ogr.Geometry)

    def test_border_geojson(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        expected_geojson = '{ "type": "Polygon", "coordinates": [ [ [ 25.0, 72.0 ],' \
                           ' [ 26.0, 72.0 ], [ 27.0, 72.0 ], [ 28.0, 72.0 ], [ 29.0, 72.0 ], ' \
                           '[ 30.0, 72.0 ], [ 31.0, 72.0 ], [ 32.0, 72.0 ], [ 33.0, 72.0 ],' \
                           ' [ 34.0, 72.0 ], [ 35.0, 72.0 ], [ 35.0, 72.0 ], [ 35.0, 71.8 ], ' \
                           '[ 35.0, 71.6 ], [ 35.0, 71.4 ], [ 35.0, 71.2 ], [ 35.0, 71.0 ],' \
                           ' [ 35.0, 70.8 ], [ 35.0, 70.6 ], [ 35.0, 70.4 ], [ 35.0, 70.2 ], ' \
                           '[ 35.0, 70.0 ], [ 35.0, 70.0 ], [ 34.0, 70.0 ], [ 33.0, 70.0 ], ' \
                           '[ 32.0, 70.0 ], [ 31.0, 70.0 ], [ 30.0, 70.0 ], [ 29.0, 70.0 ], ' \
                           '[ 28.0, 70.0 ], [ 27.0, 70.0 ], [ 26.0, 70.0 ], [ 25.0, 70.0 ], ' \
                           '[ 25.0, 70.0 ], [ 25.0, 70.2 ], [ 25.0, 70.4 ], [ 25.0, 70.6 ],' \
                           ' [ 25.0, 70.8 ], [ 25.0, 71.0 ], [ 25.0, 71.2 ], [ 25.0, 71.4 ], ' \
                           '[ 25.0, 71.6 ], [ 25.0, 71.8 ], [ 25.0, 72.0 ] ] ] }'
        border_geojson = d.get_border_geojson()
        self.assertIsInstance(border_geojson, str)
        self.assertEqual(border_geojson, expected_geojson)

    def test_overlaps_intersects_and_contains(self):
        Bergen = Domain(4326, "-te 5 60 6 61 -ts 500 500")
        WestCoast = Domain(4326, "-te 1 58 6 64 -ts 500 500")
        Norway = Domain(4326, "-te 3 55 30 72 -ts 500 500")
        Paris = Domain(4326, "-te 2 48 3 49 -ts 500 500")

        self.assertFalse(Bergen.overlaps(Norway))
        self.assertFalse(Norway.overlaps(Bergen))
        self.assertTrue(Norway.overlaps(WestCoast))
        self.assertFalse(Paris.overlaps(Norway))

        self.assertTrue(Bergen.intersects(Norway))
        self.assertTrue(Norway.intersects(Bergen))
        self.assertTrue(Norway.intersects(WestCoast))
        self.assertFalse(Paris.intersects(Norway))

        self.assertTrue(Norway.contains(Bergen))
        self.assertFalse(Bergen.contains(Norway))
        self.assertFalse(Norway.contains(WestCoast)) # why false?
        self.assertFalse(Paris.contains(Norway))


    def test_contains(self):
        Bergen = Domain(4326, EXTENT_BERGEN)
        WestCoast = Domain(4326, EXTENT_WESTCOAST)
        Norway = Domain(4326, EXTENT_NORWAY)
        Paris = Domain(4326, EXTENT_PARIS)
        self.assertTrue(Norway.contains(Bergen))
        self.assertFalse(Bergen.contains(Norway))
        self.assertFalse(Norway.contains(WestCoast))
        self.assertFalse(Paris.contains(Norway))

    def test_get_border_postgis(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        result = d.get_border_postgis()
        self.assertIsInstance(result, str)
        self.assertEquals(result, "PolygonFromText('POLYGON((25.0 72.0,26.0 72.0,27.0 72.0,28.0 "
                                  "72.0,29.0 72.0,30.0 72.0,31.0 72.0,32.0 72.0,33.0 72.0,34.0 "
                                  "72.0,35.0 72.0,35.0 72.0,35.0 71.8,35.0 71.6,35.0 71.4,35.0 "
                                  "71.2,35.0 71.0,35.0 70.8,35.0 70.6,35.0 70.4,35.0 70.2,35.0 "
                                  "70.0,35.0 70.0,34.0 70.0,33.0 70.0,32.0 70.0,31.0 70.0,30.0 "
                                  "70.0,29.0 70.0,28.0 70.0,27.0 70.0,26.0 70.0,25.0 70.0,25.0 "
                                  "70.0,25.0 70.2,25.0 70.4,25.0 70.6,25.0 70.8,25.0 71.0,25.0 "
                                  "71.2,25.0 71.4,25.0 71.6,25.0 71.8,25.0 72.0))')")

    def test_get_corners(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        lon, lat = d.get_corners()
        self.assertTrue(all(lon - [25., 25., 35., 35.] < 0.01))
        self.assertTrue(all(lat - [72., 70., 72., 70.] < 0.01))

    def test_get_min_max_lon_lat(self):
        dom = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        result = dom.get_min_max_lon_lat()
        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 4)
        for el in result:
            self.assertIsInstance(el, float)
        self.assertLess(result[0], result[1])
        self.assertLess(result[2], result[3])
        self.assertEqual(result, (25.0, 34.980000000000004, 70.004000000000005, 72.0))

    def test_get_pixelsize_meters(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        x, y = d.get_pixelsize_meters()
        self.assertEqual(int(x), 444)
        self.assertEqual(int(y), 723)
        d = Domain(ds=gdal.Open(self.test_file_projected))
        x, y = d.get_pixelsize_meters()
        self.assertEqual(int(x), 500)
        self.assertEqual(int(y), 500)

    def test_get_geotransform(self):
        input_1 = {'te': [25.0, 70.0, 35.0, 72.0], 'ts': [500.0, 500.0]}
        test_1 = ([25.0, 0.02, 0.0, 72.0, 0.0, -0.004], 500, 500)
        result = Domain._get_geotransform(input_1)
        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 3)
        self.assertIsInstance(result[0], list)
        self.assertEqual(len(result[0]), 6)
        for el in result[0]:
            self.assertIsInstance(el, float)
        self.assertIsInstance(result[1], int)
        self.assertIsInstance(result[2], int)
        self.assertEqual(result, test_1)

    def test_transform_tr(self):
        result = Domain._transform_tr(4.0, 1.3, [0.015, 0.005])
        self.assertIsInstance(result, tuple)
        self.assertEquals(len(result), 4)
        map(lambda el: self.assertIsInstance(el, float), result)
        with self.assertRaises(ValueError) as param_err:
            Domain._transform_tr(4.0, 1.3, [5.0, 0.005])
            self.assertEqual(param_err.message,
                             '"-tr" is too large. width is 4.0, height is 1.3 ')

    def test_transform_ts2(self):
        result = Domain._transform_ts(1.5, 1.0, [750.0, 500.0])
        self.assertIsInstance(result, tuple)
        self.assertEquals(len(result), 4)
        map(lambda el: self.assertIsInstance(el, float), result)

    def test_transform_points(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        lon, lat = d.transform_points([1, 2, 3], [1, 2, 3])
        self.assertEqual(type(lon), np.ndarray)
        self.assertEqual(type(lat), np.ndarray)

    def test_transform_points_inverse(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        x, y = d.transform_points([25, 26, 27], [70, 71, 72], 1)
        self.assertTrue(all(np.round(x) == [0, 50, 100]))
        self.assertTrue(all(np.round(y) == [500, 250, 0]))

    def test_transform_points_dstsrs(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        lon, lat = d.transform_points([1, 2, 3], [1, 2, 3],
        dst_srs=NSR('+proj=stere +datum=WGS84 +ellps=WGS84 +lat_0=75 +lon_0=10 +no_defs'))
        self.assertEqual(type(lon), np.ndarray)
        self.assertEqual(type(lat), np.ndarray)

    @patch.object(Domain, 'get_geolocation_grids',
                  return_value=(np.array([[-4.        , -0.66666667,  2.66666667],
                                          [-4.        , -0.66666667,  2.66666667]]),
                                np.array([[ 7.,  7.,  7.],
                                          [ 1.,  1.,  1.]])))
    def test_azimuth_y(self, mock_get_geolocation_grids):
        d = Domain(4326, "-te -4 -5 +6 +7 -ts 3 2")
        self.assertTrue((d.azimuth_y()==np.array([[ 0.,  0.,  0.], [ 0.,  0.,  0.]])).all())

    def test_shape(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        self.assertEqual(d.shape(), (500, 500))

    def test_reproject_gcps(self):
        ds = gdal.Open(self.test_file)
        d = Domain(ds=ds)
        d.reproject_gcps('+proj=stere +datum=WGS84 +ellps=WGS84 +lat_0=75 +lon_0=10 +no_defs')
        gcp = d.vrt.dataset.GetGCPs()[0]
        self.assertTrue(gcp.GCPX > 636161)
        self.assertTrue(gcp.GCPY < -288344)

    def test_reproject_gcps_auto(self):
        ds = gdal.Open(self.test_file)
        d = Domain(ds=ds)
        d.reproject_gcps()
        gcpproj = NSR(d.vrt.dataset.GetGCPProjection())
        self.assertEqual(gcpproj.GetAttrValue('PROJECTION'),
                        'Stereographic')

    @patch.multiple(Domain, get_border_geometry=DEFAULT, __init__ = Mock(return_value=None))
    def test_intersects(self, get_border_geometry):
        other_domain = MagicMock()
        d = Domain()
        d.intersects(other_domain)
        d.get_border_geometry.assert_called_once()
        d.get_border_geometry().Intersects.assert_called_once_with(other_domain.get_border_geometry())

    @patch.multiple(Domain, get_border_geometry=DEFAULT, __init__ = Mock(return_value=None))
    def test_overlaps(self, get_border_geometry):
        other_domain = MagicMock()
        d = Domain()
        d.overlaps(other_domain)
        d.get_border_geometry.assert_called_once()
        d.get_border_geometry().Overlaps.assert_called_once_with(other_domain.get_border_geometry())

    def test_get_border_dateline(self):
        dom = Domain('+proj=stere +datum=WGS84 +ellps=WGS84 +lat_0=0 +lon_0=-179 +no_defs',
                     '-te -1000000 -1000 1000000 1000 -ts 10 10')
        brd = dom.get_border()
        self.assertTrue(np.all(brd[0] > 0))
        brd = dom.get_border(fix_lon=False)
        self.assertTrue(brd[0].min() < 0)

if __name__ == "__main__":
    unittest.main()



# Name:     mapper_s1a_l2.py
# Purpose:      Mapping for Sentinel-1 level-2 data
#
# Author:       Morten Wergeland Hansen
# Modified: Morten Wergeland Hansen
#
# Created:  13.03.2014
# Last modified:17.07.2014 11:57
# Copyright:    (c) NERSC
# License:      This file is part of NANSAT. NANSAT is free software: you can
#               redistribute it and/or modify it under the terms of the GNU
#               General Public License as published by the Free Software
#               Foundation, version 3 of the License.
#               http://www.gnu.org/licenses/gpl-3.0.html This program is
#               distributed in the hope that it will be useful, but WITHOUT ANY
#               WARRANTY without even the implied warranty of MERCHANTABILITY
#               or FITNESS FOR A PARTICULAR PURPOSE.
import os
from dateutil.parser import parse

import gdal

from nansat.vrt import VRT
from nansat.exceptions import WrongMapperError
from nansat.nsr import NSR


class Mapper(VRT):
    '''
        Create VRT with mapping of Sentinel-1A stripmap mode (S1A_SM)
    '''

    def __init__(self, filename, gdalDataset, gdalMetadata,
                 product_type='RVL', GCP_COUNT=10, **kwargs):
        '''
        Parameters
        ----------
        product_type: string
            Sentinel-1 level-2 ocean product type/component, i.e. ocean swell
            spectra (OSW), ocean wind field (OWI), or radial surface velocity
            (RVL) (RVL is the default)
        GCP_COUNT : int
            number of GCPs along each dimention
        '''
        fPathName, fExt = os.path.splitext(filename)

        # List of Sentinel-1 level-2 components
        unwanted_product_components = ['osw', 'owi', 'rvl']
        # Remove requested 'product_type' from list of unwanted
        unwanted_product_components.pop(unwanted_product_components.index(
                                        product_type.lower()))

        # Check if it is Sentinel-1 (or ASAR) level-2 (in S1 data format)
        if not gdalMetadata or not 'NC_GLOBAL' in gdalMetadata.keys():
            raise WrongMapperError(filename)
        else:
            title = gdalMetadata['NC_GLOBAL#TITLE']

        # Raise error if it is not Sentinel-1 format
        if not 'Sentinel-1' or 'ASA' in title:
            raise WrongMapperError(filename)

        metadata = {}
        for key, val in gdalMetadata.iteritems():
            new_key = key.split('#')[-1]
            metadata[new_key] = val

        subDatasets = gdalDataset.GetSubDatasets()
        filenames = [f[0] for f in subDatasets]

        rm_bands = []
        # Find all data that is not relevant for the selected product type
        # and get bands of longitude, latitude and zero doppler time
        for i, f in enumerate(filenames):
            if f.split(':')[-1][:3] in unwanted_product_components:
                rm_bands.append(i)
            if 'Lon' in f.split(':')[-1]:
                lon_ds = gdal.Open(f)
                rm_bands.append(i)
            if 'Lat' in f.split(':')[-1]:
                lat_ds = gdal.Open(f)
                rm_bands.append(i)
            if 'ZeroDopplerTime' in f.split(':')[-1]:
                zdt_ds = gdal.Open(f)
                rm_bands.append(i)
        # Remove bands in rm_bands from the list of bands to add to the Nansat
        # object
        filenames = [f for i, f in enumerate(filenames) if not i in rm_bands]
           #     (
           # 'Lon' in f.split(':')[-1] or
           # 'Lat' in f.split(':')[-1] or
           # 'ZeroDopplerTime' in f.split(':')[-1] )]

        # create empty VRT dataset
        self._init_from_gdal_dataset(gdal.Open(subDatasets[0][0]), metadata=metadata)

        # The zero Doppler time grid is 3-dimensional - the last dimension is a
        # char array with the time as year, month, day, etc.
        # Will not bother with it yet...
        #for iBand in range(zdt_ds.RasterCount):
        #    subBand = zdt_ds.GetRasterBand(iBand+1)

        XSize = lon_ds.RasterXSize
        YSize = lon_ds.RasterYSize

        # get projection from the lon and lat datasets
        longitude = lon_ds.ReadAsArray()
        latitude = lat_ds.ReadAsArray()
        # estimate step of GCPs
        step0 = max(1, int(float(latitude.shape[0]) / GCP_COUNT))
        step1 = max(1, int(float(latitude.shape[1]) / GCP_COUNT))
        self.logger.debug('gcpCount: >%s<, %d %d %f %d %d',
                          title,
                          latitude.shape[0], latitude.shape[1],
                          GCP_COUNT, step0, step1)

        # estimate pixel/line step of the geolocation arrays
        pixelStep = 1
        lineStep = 1
        self.logger.debug('pixel/lineStep %f %f' % (pixelStep, lineStep))
        # generate list of GCPs
        dx = .5
        dy = .5
        gcps = []
        k = 0
        for i0 in range(0, latitude.shape[0], step0):
            for i1 in range(0, latitude.shape[1], step1):
                # create GCP with X,Y,pixel,line from lat/lon matrices
                lon = float(longitude[i0, i1])
                lat = float(latitude[i0, i1])
                if (lon >= -180 and lon <= 180 and lat >= -90 and lat <= 90):
                    gcp = gdal.GCP(lon, lat, 0, i1 * pixelStep + dx,
                                   i0 * lineStep + dy)
                    self.logger.debug('%d %d %d %f %f',
                                      k, gcp.GCPPixel, gcp.GCPLine,
                                      gcp.GCPX, gcp.GCPY)
                    gcps.append(gcp)
                    k += 1

        # append GCPs and lat/lon projection to the vsiDataset
        self.dataset.SetGCPs(gcps, NSR().wkt)

        # define band specific parameters
        metaDict = []
        geoFileDict = {}
        xDatasetSource = ''
        yDatasetSource = ''
        for i, filename in enumerate(filenames):
            band = gdal.Open(filename)

            # check that the band size is the same size as the latitude and
            # longitude grids
            if (band.RasterXSize != XSize or
                    band.RasterYSize != YSize):
                raise IndexError(('Size of sub-dataset is different from size '
                                  'of longitude and latitude grids'))

            bandMetadata = band.GetMetadata()
            # generate src metadata
            src = {'SourceFilename': filename,
                   'SourceBand': 1
                   }

            # Generate dst metadata
            short_name = filename.split(':')[-1]
            dst = {'name': short_name,
                   'short_name': short_name,
                   'long_name': bandMetadata[short_name+'#long_name'],
                   'units': bandMetadata[short_name+'#units'],
                   #'wkv': ,
                   }

            # append band with src and dst dictionaries
            metaDict.append({'src': src, 'dst': dst})

        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)

        metaDict = []
        for i in range(self.dataset.RasterCount):
            if 'Nrcs' in self.dataset.GetRasterBand(i+1).GetMetadata()['name']:
                metaDict.append({
                    'src': {'SourceFilename': (self.dataset.GetRasterBand(i+1).
                                GetMetadata()['SourceFilename']),
                            'SourceBand': 1
                            },
                    'dst': {
                        'short_name': 'sigma0',
                        'wkv': 'surface_backwards_scattering_coefficient_of_radar_wave',
                        'PixelFunctionType':  'dB2pow',
                        'polarization': (self.dataset.
                                         GetMetadata()['POLARISATION']),
                        'suffix': self.dataset.GetMetadata()['POLARISATION'],
                        'dataType': 6,
                    }
                })

        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)

        # set time
        self.dataset.SetMetadataItem('time_coverage_start',
            parse(self.dataset.GetMetadata()['SOURCE_ACQUISITION_UTC_TIME']).isoformat())



# Name:    figure.py
# Purpose: Container of Figure class
# Authors:      Asuka Yamakawa, Anton Korosov, Knut-Frode Dagestad,
#               Morten W. Hansen, Alexander Myasoyedov,
#               Dmitry Petrenko, Evgeny Morozov
# Created:      29.06.2011
# Copyright:    (c) NERSC 2011 - 2013
# Licence:
# This file is part of NANSAT.
# NANSAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
# http://www.gnu.org/licenses/gpl-3.0.html
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
from __future__ import absolute_import

import os
from math import floor, log10

import numpy as np
try:
    if 'DISPLAY' not in os.environ:
        import matplotlib; matplotlib.use('Agg')
    from matplotlib import cm
except ImportError:
    MATPLOT_LIB_IS_INSTALLED = False
else:
    MATPLOT_LIB_IS_INSTALLED = True

try:
    import Image
    import ImageDraw
    import ImageFont
except:
    from PIL import Image, ImageDraw, ImageFont

from nansat.utils import add_logger


class Figure(object):
    """Perform operations with graphical files: create, append legend, save.

    Figure instance is created in the Nansat.write_figure method.
    The methods below are applied consequently in order to generate a figure
    from one or three bands, estimate min/max, apply logarithmic scaling,
    convert to uint8, append legend, save to a file

    **Modifies:** self.sizeX, self.sizeY (int), width and height of the image

    **Modifies:** self.pilImg (PIL image), figure

    **Modifies:** self.pilImgLegend (PIL image)

    Note
    ----
    If pilImgLegend is None, legend is not added to the figure.
    If it is replaced, pilImgLegend includes text string, color-bar,
    longName and units.

    Parameters
    -----------
    array : numpy array (2D or 3D)
        dataset from Nansat

    cmin : number (int ot float) or [number, number, number]
        0, minimum value of varibale in the matrix to be shown
    cmax : number (int ot float) or [number, number, number]
        1, minimum value of varibale in the matrix to be shown
    gamma : float, >0
        2, coefficient for tone curve udjustment
    subsetArraySize : int
        100000, size of the subset array which is used to get histogram
    numOfColor : int
        250, number of colors for use of the palette.
        254th is black and 255th is white.
    cmapName : string
        'jet', name of Matplotlib colormaps
        see --> http://www.scipy.org/Cookbook/Matplotlib/Show_colormaps
    ratio : float, [0 1]
        1.0, ratio of pixels which are used to write the figure
    numOfTicks : int
        5, number of ticks on a colorbar
    titleString : string
        '', title of legend (1st line)
    caption : string
        '', caption of the legend (2nd line, e.g. long name and units)
    fontRatio : positive float
        1, factor for changing the fontSize.
    fontSize : int
        12, size of the font of title, caption and ticks.
        If not given, fontSize is calculated using fontRatio:
        fontSize = height / 45 * fontRatio.
        fontSize has priority over fontRatio
    logarithm : boolean, defult = False
        If True, tone curve is used to convert pixel values.
        If False, linear.
    legend : boolean, default = False
        if True, information as textString, colorbar, longName and
        units are added in the figure.
    mask_array : 2D numpy array, int, the shape should be equal to
        array.shape. If given, this array is used for masking land,
        clouds, etc on the output image. Value of the array are
        indices. LUT from mask_lut is used for coloring upon this
        indices.
    mask_lut : dictionary
        Look-Up-Table with colors for masking land, clouds etc. Used
        tgether with mask_array:
        {0, [0,0,0], 1, [100,100,100], 2: [150,150,150], 3: [0,0,255]}
        index

            - 0 - will have black color
            - 1 - dark gray
            - 2 - light gray
            - 3 - blue

    logoFileName : string
        name of the file with logo
    logoLocation : list of two int, default = [0,0]
        X and Y offset of the image
        If positive - offset is from left, upper edge
        If Negative - from right, lower edge
        Offset is calculated from the entire image legend inclusive
    logoSize : list of two int
        desired X,Y size of logo. If None - original size is used
    latGrid : numpy array
        full size array with latitudes. For adding lat/lon grid lines
    lonGrid : numpy array
        full size array with longitudes. For adding lat/lon grid lines
    nGridLines : int
        number of lat/lon grid lines to show
    latlonLabels : int
        number of lat/lon labels to show along each side.
    transparency : int
        transparency of the image background(mask), set for PIL alpha
        mask in Figure.save()
    default : None

    Other parameters
    ----------------
    LEGEND_HEIGHT : float, [0 1]
        0.1, legend height relative to image height
    CBAR_HEIGHTMIN : int
        5, minimum colorbar height, pixels
    CBAR_HEIGHT : float, [0 1]
        0.15,  colorbar height relative to image height
    CBAR_WIDTH : float [0 1]
        0.8, colorbar width  relative to legend width
    CBAR_LOCATION_X : float [0 1]
        0.1, colorbar offset X  relative to legend width
    CBAR_LOCATION_Y : float [0 1]
        0.5,  colorbar offset Y  relative to legend height
    CBTICK_LOC_ADJUST_X : int
        5,  colorbar tick label offset X, pixels
    CBTICK_LOC_ADJUST_Y : int
        3,  colorbar tick label offset Y, pixels
    CAPTION_LOCATION_X : float, [0 1]
        0.1, caption offset X relative to legend width
    CAPTION_LOCATION_Y : float, [0 1]
        0.1, caption offset Y relative to legend height
    TITLE_LOCATION_X : float, [0 1]
        0.1, title offset X relative to legend width
    TITLE_LOCATION_Y :
        0.3, title  offset Y relative to legend height
    DEFAULT_EXTENSION : string
        '.png'

    """

    # default values of ALL params of Figure
    cmin = [0.]
    cmax = [1.]
    gamma = 2.
    subsetArraySize = 100000
    numOfColor = 250
    cmapName = 'jet'
    ratio = 1.0
    numOfTicks = 5
    titleString = ''
    caption = ''
    fontRatio = 1
    fontSize = None
    logarithm = False
    legend = False
    mask_array = None
    mask_lut = None

    logoFileName = None
    logoLocation = [0, 0]
    logoSize = None

    latGrid = None
    lonGrid = None
    lonTicks = 5
    latTicks = 5

    transparency = None

    LEGEND_HEIGHT = 0.1
    CBAR_HEIGHTMIN = 5
    CBAR_HEIGHT = 0.15
    CBAR_WIDTH = 0.8
    CBAR_LOCATION_X = 0.1
    CBAR_LOCATION_Y = 0.5
    CBTICK_LOC_ADJUST_X = 5
    CBTICK_LOC_ADJUST_Y = 3
    CAPTION_LOCATION_X = 0.1
    CAPTION_LOCATION_Y = 0.25
    TITLE_LOCATION_X = 0.1
    TITLE_LOCATION_Y = 0.05
    DEFAULT_EXTENSION = '.png'

    palette = None
    pilImg = None
    pilImgLegend = None

    extensionList = ['png', 'PNG', 'tif', 'TIF', 'bmp',
                     'BMP', 'jpg', 'JPG', 'jpeg', 'JPEG']

    _cmapName = 'jet'

    # instance attributes
    array = None
    def __init__(self, nparray, **kwargs):
        """ Set attributes

        See class Figure(object) for information about:

        Modifies
        --------

        Parameters
        ----------

        Advanced parameters
        --------------------

        """
        # make a copy of nparray (otherwise a new reference to the same data is
        # created and the original input data is destroyed at process())
        array = np.array(nparray)

        self.logger = add_logger('Nansat')

        # if 2D array is given, reshape to 3D
        if array.ndim == 2:
            self.array = array.reshape(1, array.shape[0], array.shape[1])
        else:
            self.array = array

        # note swaping of axis by PIL
        self.width = self.array.shape[2]
        self.height = self.array.shape[1]

        # modify the default values using input values
        self._set_defaults(kwargs)

        # set fonts for Legend
        self.fontFileName = os.path.join(os.path.dirname(
                                         os.path.realpath(__file__)),
                                         'fonts/DejaVuSans.ttf')

    def apply_logarithm(self, **kwargs):
        """Apply a tone curve to the array

        After the normalization of the values from 0 to 1, logarithm is applied
        Then the values are converted to the normal scale.

        **Modifies**: self.array (numpy array)

        Parameters
        -----------
        **kwargs : dict
            Any of Figure parameters

        """
        # modify default parameters
        self._set_defaults(kwargs)

        # apply logarithm/gamme correction to pixel values
        for iBand in range(self.array.shape[0]):
            self.array[iBand, :, :] = (
                (np.power((self.array[iBand, :, :] - self.cmin[iBand]) /
                         (self.cmax[iBand] - self.cmin[iBand]),
                          (1.0 / self.gamma))) *
                (self.cmax[iBand] - self.cmin[iBand]) +
                self.cmin[iBand])

    def apply_mask(self, **kwargs):
        """Apply mask for coloring land, clouds, etc

        If mask_array and mask_lut are provided as input parameters.
        The pixels in self.array which have index equal to mask_lut key
        in mask_array will have color equal to mask_lut value.

        **Modifies:** self.array (numpy array)

        Note
        ----
        apply_mask should be called only after convert_palettesize
        (i.e. to uint8 data)

        Parameters
        -----------
        **kwargs : dict
            Any of Figure parameters

        """
        # modify default parameters
        self._set_defaults(kwargs)

        # get values of free indices in the palette
        availIndices = range(self.numOfColor, 255 - 1)

        # for all lut color indices
        for i, maskValue in enumerate(self.mask_lut):
            if i < len(availIndices):
                # get color for that index
                maskColor = self.mask_lut[maskValue]
                # get indices for that index
                maskIndices = self.mask_array == maskValue
                # exchange colors
                if self.array.shape[0] == 1:
                    # in a indexed image
                    self.array[0][maskIndices] = availIndices[i]
                elif self.array.shape[0] == 3:
                    # in RGB image
                    for c in range(0, 3):
                        self.array[c][maskIndices] = maskColor[c]

                # exchange palette
                self.palette[(availIndices[i] * 3):
                             (availIndices[i] * 3 + 3)] = maskColor

    def add_logo(self, **kwargs):
        """Insert logo into the PIL image

        Read logo from file as PIL.
        Resize to the given size.
        Pan using the given location.
        Paste into pilImg.

        **Modifies:** self.pilImg (PIL image)

        Parameters
        ----------
        **kwargs : dict
            Any of Figure parameters

        """
        # set/get default parameters
        self._set_defaults(kwargs)
        logoFileName = self.logoFileName
        logoLocation = self.logoLocation
        logoSize = self.logoSize

        # check if pilImg was created already
        if self.pilImg is None:
            self.logger.warning('Create PIL image first')
            return
        # check if file is available
        try:
            logoImg = Image.open(logoFileName)
        except:
            self.logger.warning('No logo file %s' % logoFileName)
            return
        # resize if required
        if logoSize is None:
            logoSize = logoImg.size
        else:
            logoImg = logoImg.resize(logoSize)
        # get location of the logo w.r.t. sign of logoLocation
        box = [0, 0, logoSize[0], logoSize[1]]
        for dim in range(2):
            if logoLocation[dim] >= 0:
                box[dim + 0] = box[dim + 0] + logoLocation[dim + 0]
                box[dim + 2] = box[dim + 2] + logoLocation[dim + 0]
            else:
                box[dim + 0] = (self.pilImg.size[dim + 0] +
                                logoLocation[dim + 0] -
                                logoSize[dim + 0])
                box[dim + 2] = (self.pilImg.size[dim + 0] +
                                logoLocation[dim + 0])

        self.pilImg = self.pilImg.convert('RGB')
        self.pilImg.paste(logoImg, tuple(box))

    def add_latlon_grids(self, **kwargs):
        """Add lat/lon grid lines into the PIL image

        Compute step of the grid.
        Make matrices with binarized lat/lon.
        Find edge (make line).
        Convert to mask.
        Add mask to PIL

        **Modifies:** self.pilImg (PIL image), added lat/lon grid lines

        Parameters
        ----------

        latGrid : numpy array
            array with values of latitudes
        lonGrid : numpy array
            array with values of longitudes
        lonTicks : int or list
            number of lines to draw
            or locations of gridlines
        latTicks : int or list
            number of lines to draw
            or locations of gridlines
        **kwargs : dict
            any of Figure parameters

        """
        # modify default values
        self._set_defaults(kwargs)

        # test availability of grids
        if (self.latGrid is None or self.lonGrid is None):
            return

        # get vectors with ticks based on input
        latTicks = self._get_auto_ticks(self.latTicks, self.latGrid)
        lonTicks = self._get_auto_ticks(self.lonTicks, self.lonGrid)

        # convert lat/lon grids to indices
        latI = np.zeros(self.latGrid.shape, 'int8')
        lonI = np.zeros(self.latGrid.shape, 'int8')
        for latTick in latTicks:
            latI[self.latGrid > latTick] += 1
        for lonTick in lonTicks:
            lonI[self.lonGrid > lonTick] += 1

        # find pixels on the grid lines (binarize)
        latI = np.sum(np.gradient(latI), axis=0)
        lonI = np.sum(np.gradient(lonI), axis=0)

        # Set border pixels equal to nearest neighbouring pixels - the error
        # should be minor (alternatively, they should all be 0) - this and the
        # two lines above solve VisibleDeprecationWarning in numpy<1.13.0 and
        # IndexError in numpy>=1.13.0
        latI[latI.shape[0]-1, :] = latI[latI.shape[0]-2, :]
        latI[:, latI.shape[0]-1] = latI[:, latI.shape[0]-2]
        lonI[lonI.shape[0]-1, :] = lonI[lonI.shape[0]-2, :]
        lonI[:, lonI.shape[0]-1] = lonI[:, lonI.shape[0]-2]

        # make grid from both lat and lon
        latI += lonI
        latI[latI != 0] = 1

        # add mask to the image
        self.apply_mask(mask_array=latI, mask_lut={1: [255, 255, 255]})

    def _get_auto_ticks(self, ticks, grid):
        """Automatically create a list of lon or lat ticks from number of list

        Parameters
        ----------
            ticks : int or list
                number or location of ticks
            grid : ndarray
                grid with lon or lat
        Returns
        -------
            ticks : list
                location of ticks

        """
        gridMin = grid.min()
        gridMax = grid.max()

        if type(ticks) is int:
            ticks = np.linspace(gridMin, gridMax, ticks)
        elif type(ticks) in [list, tuple]:
            newTicks = []
            for tick in ticks:
                if tick >= gridMin and tick <= gridMax:
                    newTicks.append(tick)
            ticks = newTicks
        else:
            raise ValueError('Incorrect type of ticks')

        return ticks

    def add_latlon_labels(self, **kwargs):
        """Add lat/lon labels along upper and left side

        Compute step of lables.
        Get lat/lon for these labels from latGrid, lonGrid
        Print lables to PIL in white.

        **Modifies:** self.pilImg (PIL image), added lat/lon labels

        Parameters
        ----------
        latGrid : numpy array
            array with values of latitudes
        lonGrid : numpy array
            array with values of longitudes
        lonTicks : int or list
            number of lines to draw
            or locations of gridlines
        latTicks : int or list
            number of lines to draw
            or locations of gridlines
        **kwargs : dict
            Any of Figure parameters

        """
        # modify default values
        self._set_defaults(kwargs)

        # test availability of grids
        if (self.latGrid is None or self.lonGrid is None):
            return

        draw = ImageDraw.Draw(self.pilImg)
        font = ImageFont.truetype(self.fontFileName, self.fontSize)

        # get vectors with ticks based on input
        latTicks = self._get_auto_ticks(self.latTicks, self.latGrid)
        lonTicks = self._get_auto_ticks(self.lonTicks, self.lonGrid)

        # get corresponding lons from upper edge and lats from left edge
        lonTicksIdx = self._get_tick_index_from_grid(lonTicks, self.lonGrid,
                                       1, self.lonGrid.shape[1])
        latTicksIdx = self._get_tick_index_from_grid(latTicks, self.latGrid,
                                       self.lonGrid.shape[0], 1)

        # draw lons
        lonsOffset = self.lonGrid.shape[1] / len(lonTicksIdx) / 8.
        for lonTickIdx in lonTicksIdx:
            lon = self.lonGrid[0, lonTickIdx]
            draw.text((lonTickIdx+lonsOffset, 0), '%4.2f' % lon,
                                                    fill=255, font=font)

        # draw lats
        latsOffset = self.latGrid.shape[0] / len(latTicksIdx) / 8.
        for latTickIdx in latTicksIdx:
            lat = self.latGrid[latTickIdx, 0]
            draw.text((0, latTickIdx+latsOffset), '%4.2f' % lat,
                      fill=255, font=font)

    def _get_tick_index_from_grid(self, ticks, grid, rows, cols):
        """Get index of pixels from lon/lat grids closest given ticks

        Parameters
        ----------
        ticks : int or list
            number or location of ticks
        grid : ndarray
            grid with lon or lat
        rows : int
            from which rows to return pixels
        cols : int
            from which cols to return pixels

        Returns
        -------
        ticks : list
            index of ticks

        """
        newTicksIdx = []
        for tick in ticks:
            diff = np.abs(grid[:rows, :cols] - tick).flatten()
            minDiffIdx = np.nonzero(diff == diff.min())[0][0]
            if minDiffIdx > 0:
                newTicksIdx.append(minDiffIdx)
        return newTicksIdx

    def clim_from_histogram(self, **kwargs):
        """Estimate min and max pixel values from histogram

        if ratio=1.0, simply the minimum and maximum values are returned.
        if 0 < ratio < 1.0, get the histogram of the pixel values.
        Then get rid of (1.0-ratio)/2 from the both sides and
        return the minimum and maximum values.

        Parameters
        -----------
        **kwargs : dict
            Any of Figure parameters

        Returns
        --------
        clim : numpy array 2D ((3x2) or (1x2))
            minimum and maximum pixel values for each band

        """
        # modify default values
        self._set_defaults(kwargs)
        ratio = self.ratio

        # find masked pixels if mask_array and mask_lut provided
        masked = None
        if self.mask_array is not None and self.mask_lut is not None:
            masked = np.zeros(self.mask_array.shape, 'bool')
            for lutVal in self.mask_lut:
                masked = masked + (self.mask_array == lutVal)

        # create a ratio list for each band
        if not (isinstance(ratio, float) or isinstance(ratio, int)):
            raise ValueError('Incorrect input ratio %s' % str(ratio))

        # create a ratio list for each band
        if ratio <= 0 or ratio > 1:
            raise ValueError('Incorrect input ratio %s' % str(ratio))

        # create a 2D array and set min and max values
        clim = [[0] * self.array.shape[0], [0] * self.array.shape[0]]
        for iBand in range(self.array.shape[0]):
            bandArray = self.array[iBand, :, :]
            # remove masked data
            if masked is not None:
                bandArray = bandArray[masked == 0]
            # remove nan, inf
            bandArray = bandArray[np.isfinite(bandArray)]
            # get percentile
            percentileMin = 100 * (1 - ratio) / 2.
            percentileMax = 100 * (1 - (1 - ratio) / 2.)
            if bandArray.size > 0:
                clim[0][iBand] = np.percentile(bandArray, percentileMin)
                clim[1][iBand] = np.percentile(bandArray, percentileMax)
            else:
                clim[0][iBand], clim[1][iBand] = 0, 1

        self.color_limits = clim
        return clim

    def clip(self, **kwargs):
        """Convert self.array to values between cmin and cmax

        if pixel value < cmin, replaced to cmin.

        if pixel value > cmax, replaced to cmax.

        **Modifies:** self.array (numpy array)

        **Modifies:** self.cmin, self.cmax : allowed min/max values

        Parameters
        -----------
        **kwargs : dict
            Any of Figure parameters

        """
        # modify default parameters
        self._set_defaults(kwargs)

        for iBand in range(self.array.shape[0]):
            # if clipping integer matrix, make clipping ranges valid
            if self.array.dtype in ['int8', 'uint8', 'int16', 'uint16']:
                self.cmin[iBand] = np.ceil(self.cmin[iBand])
                self.cmin[iBand] = np.floor(self.cmin[iBand])

            # Clipping, allowing for reversed colorscale (cmin > cmax)
            clipMin = np.min([self.cmin[iBand], self.cmax[iBand]])
            clipMax = np.max([self.cmin[iBand], self.cmax[iBand]])
            self.array[iBand, :, :] = np.clip(self.array[iBand, :, :],
                                              clipMin, clipMax)

    def convert_palettesize(self, **kwargs):
        """Convert self.array to palette color size in uint8

        **Modifies:** self.array (numpy array)

        Parameters
        -----------
        **kwargs : dict
            Any of Figure parameters

        """
        # modify default values
        self._set_defaults(kwargs)

        for iBand in range(self.array.shape[0]):
            self.array[iBand, :, :] = (
                (self.array[iBand, :, :].astype('float32') -
                 self.cmin[iBand]) *
                (self.numOfColor - 1) /
                (self.cmax[iBand] - self.cmin[iBand]))

        self.array = self.array.astype(np.uint8)

    def create_legend(self, **kwargs):
        """self.legend is replaced from None to PIL image

        PIL image includes colorbar, caption, and titleString.

        **Modifies:** self.legend (PIL image)

        Parameters
        -----------
        **kwargs : dict
            Any of Figure parameters

        """
        # modify default parameters
        self._set_defaults(kwargs)

        # set fonts size for colorbar
        font = ImageFont.truetype(self.fontFileName, self.fontSize)

        # create a pilImage for the legend
        self.pilImgLegend = Image.new('P', (self.width,
                                            int(self.height *
                                                self.LEGEND_HEIGHT)), 255)
        draw = ImageDraw.Draw(self.pilImgLegend)

        # set black color
        if self.array.shape[0] == 1:
            black = 254
        else:
            black = (0, 0, 0)

        # if 1 band, draw the color bar
        if self.array.shape[0] == 1:
            # make an array for color bar
            bar = np.outer(np.ones(max(int(self.pilImgLegend.size[1] *
                           self.CBAR_HEIGHT), self.CBAR_HEIGHTMIN)),
                           np.linspace(0, self.numOfColor,
                                       int(self.pilImgLegend.size[0] *
                                           self.CBAR_WIDTH)))
            # create a colorbar pil Image
            pilImgCbar = Image.fromarray(np.uint8(bar))
            # paste the colorbar pilImage on Legend pilImage
            self.pilImgLegend.paste(pilImgCbar,
                                    (int(self.pilImgLegend.size[0] *
                                         self.CBAR_LOCATION_X),
                                     int(self.pilImgLegend.size[1] *
                                         self.CBAR_LOCATION_Y)))
            # create a scale for the colorbar
            scaleLocation = np.linspace(0, 1, self.numOfTicks)
            scaleArray = scaleLocation
            if self.logarithm:
                scaleArray = (np.power(scaleArray, (1.0 / self.gamma)))
            scaleArray = (scaleArray * (self.cmax[0] -
                          self.cmin[0]) + self.cmin[0])
            scaleArray = list(map(self._round_number, scaleArray))
            # draw scales and lines on the legend pilImage
            for iTick in range(self.numOfTicks):
                coordX = int(scaleLocation[iTick] *
                             self.pilImgLegend.size[0] *
                             self.CBAR_WIDTH +
                             int(self.pilImgLegend.size[0] *
                                 self.CBAR_LOCATION_X))

                box = (coordX, int(self.pilImgLegend.size[1] *
                                   self.CBAR_LOCATION_Y),
                       coordX, int(self.pilImgLegend.size[1] *
                                  (self.CBAR_LOCATION_Y +
                                   self.CBAR_HEIGHT)) - 1)
                draw.line(box, fill=black)
                box = (coordX + self.CBTICK_LOC_ADJUST_X,
                       int(self.pilImgLegend.size[1] *
                           (self.CBAR_LOCATION_Y +
                            self.CBAR_HEIGHT)) +
                       self.CBTICK_LOC_ADJUST_Y)
                draw.text(box, scaleArray[iTick], fill=black, font=font)

        # draw longname and units
        box = (int(self.pilImgLegend.size[0] * self.CAPTION_LOCATION_X),
               int(self.pilImgLegend.size[1] * self.CAPTION_LOCATION_Y))
        draw.text(box, str(self.caption), fill=black, font=font)

        # if titleString is given, draw it
        if self.titleString != '':
            # write text each line onto pilImgCanvas
            textHeight = int(self.pilImgLegend.size[1] *
                             self.TITLE_LOCATION_Y)
            for line in self.titleString.splitlines():
                draw.text((int(self.pilImgLegend.size[0] *
                               self.TITLE_LOCATION_X),
                           textHeight), line, fill=black, font=font)
                text = draw.textsize(line, font=font)
                textHeight += text[1]

    def create_pilImage(self, **kwargs):
        """self.create_pilImage is replaced from None to PIL image

        If three images are given, create a image with RGB mode.
        if self.pilImgLegend is not None, it is pasted.

        If one image is given, create a image with P(palette) mode.
        if self.pilImgLegend is not None,
        self.array is extended before create the pilImag and
        then paste pilImgLegend onto it.

        **Modifies:** self.pilImg (PIL image), PIL image with / without the legend

        **Modifies:** self.array (replace to None)

        Parameters
        -----------
        **kwargs : dict
            Any of Figure parameters

        """
        # modify default parameters
        self._set_defaults(kwargs)

        # if legend is created, expand array with empty space below the data
        if self.pilImgLegend is not None:
            appendArray = 255 * np.ones((self.array.shape[0],
                                         self.pilImgLegend.size[1],
                                         self.width), 'uint8')
            self.array = np.append(self.array, appendArray, 1)

        # create a new PIL image from three bands (RGB) or from one (palette)
        if self.array.shape[0] == 3:
            self.pilImg = Image.merge('RGB',
                                      (Image.fromarray(self.array[0, :, :]),
                                       Image.fromarray(self.array[1, :, :]),
                                       Image.fromarray(self.array[2, :, :])))
        else:
            self.pilImg = Image.fromarray(self.array[0, :, :])
            self.pilImg.putpalette(self.palette)

        # append legend
        if self.pilImgLegend is not None:
            self.pilImg.paste(self.pilImgLegend, (0, self.height))

    def process(self, **kwargs):
        """Do all common operations for preparation of a figure for saving

        #. Modify default values of parameters by the provided ones (if any)
        #. Clip to min/max
        #. Apply logarithm if required
        #. Convert data to uint8
        #. Create palette
        #. Apply mask for colouring land, clouds, etc if required
        #. Create legend if required
        #. Create PIL image
        #. Add logo if required

        **Modifies:** self.d

        **Modifies:** self.array

        **Modifies:** self.palette

        **Modifies:** self.pilImgLegend

        **Modifies:** self.pilImg

        Parameters
        -----------
        **kwargs : dict
            Any of Figure parameters

        """
        # modify default parameters
        self._set_defaults(kwargs)

        # set fontSize using fontRatio if fontSize is not given at input
        if self.fontSize is None:
            self.fontSize = int(self.array.shape[1] / 45. * self.fontRatio)

        # if the image is reprojected it has 0 values
        # we replace them with mask before creating PIL Image
        self.reprojMask = self.array[0, :, :] == 0

        # clip values to min/max
        self.clip()

        # apply logarithm
        if self.logarithm:
            self.apply_logarithm()

        # convert to uint8
        self.convert_palettesize()

        # create the paletter
        self._create_palette()

        # apply colored mask (land mask, cloud mask and something else)
        if self.mask_array is not None and self.mask_lut is not None:
            self.apply_mask()

        # add lat/lon grids lines if latGrid and lonGrid are given
        self.add_latlon_grids()

        # append legend
        if self.legend:
            self.create_legend()

        # create PIL image ready for saving
        self.create_pilImage(**kwargs)

        # add labels with lats/lons
        self.add_latlon_labels()

        # add logo
        if self.logoFileName is not None:
            self.add_logo()

    def _make_transparent_color(self):
        """makes colors specified by self.transparency
        and self.reprojMask (if the image is reprojected) transparent

        **Modifies:** self.pilImg (PIL image), Adds transparency to PIL image

        """
        self.pilImg = self.pilImg.convert('RGBA')
        datas = self.pilImg.getdata()
        newData = list()

        for item in datas:
            if (item[0] == self.transparency[0] and
                    item[1] == self.transparency[1] and
                    item[2] == self.transparency[2]):
                newData.append((255, 255, 255, 0))
            else:
                newData.append(item)

        self.pilImg.putdata(newData)

        # The alphaMask is set in process() before clip() the Image
        img = np.array(self.pilImg)
        img[:, :, 3][self.reprojMask] = 0
        self.pilImg = Image.fromarray(np.uint8(img))

    def save(self, fileName, **kwargs):
        """Save self.pilImg to a physical file

        If given extension is JPG, convert the image mode from Palette to RGB.

        **Modifies:** self.pilImg (None)

        Parameters
        ----------
        fileName : string
            name of outputfile
        **kwargs : dict
            Any of Figure parameters

        """
        # modify default values
        self._set_defaults(kwargs)

        if not((fileName.split('.')[-1] in self.extensionList)):
            fileName = fileName + self.DEFAULT_EXTENSION

        fileExtension = fileName.split('.')[-1]
        if fileExtension in ['jpg', 'JPG', 'jpeg', 'JPEG']:
            self.pilImg = self.pilImg.convert('RGB')

        if self.transparency is not None:
            self._make_transparent_color()
        self.pilImg.save(fileName)

    def _create_palette(self):
        """Create a palette based on Matplotlib colormap name

        default number of color palette is 250.
        it means 6 colors are possible to use for other purposes.
        the last palette (255) is white and the second last (254) is black.

        **Modifies:** self.palette (numpy array) (uint8)

        """
        if not MATPLOT_LIB_IS_INSTALLED:
            # Make grayscale colormap
            cmap = np.vstack([np.arange(256.),
                              np.arange(256.),
                              np.arange(256.),
                              np.ones(256)*255]).T
            cmapLUT = cmap[:self.numOfColor, :]
        else:
            # test if given colormap name is in builtin or added colormaps
            try:
                cmap = cm.get_cmap(self.cmapName)
            except:
                self.logger.error('%s is not a valid colormap' % self.cmapName)
                self.cmapName = self._cmapName
            # get colormap by name
            cmap = cm.get_cmap(self.cmapName)
            # get colormap look-up
            cmapLUT = np.uint8(cmap(range(self.numOfColor)) * 255)

        # replace all last colors to black and...
        lut = np.zeros((3, 256), 'uint8')
        lut[:, :self.numOfColor] = cmapLUT.T[:3]
        # ...and the most last color to white
        lut[:, -1] = 255

        # set palette to be used by PIL
        self.palette = lut.T.flatten().astype(np.uint8)

    def _round_number(self, val):
        """Return writing format for scale on the colorbar

        Parameters
        ----------
        val : int / float / exponential

        Returns
        --------
        colorbar format : str

        """
        frmts = {-2: '%.2f', -1: '%.1f', 0: '%.2f',
                 1: '%.1f', 2: '%d', 3: '%d'}
        if val == 0:
            frmt = '%d'
        else:
            digit = floor(log10(abs(val)))
            if digit in frmts:
                frmt = frmts[digit]
            else:
                frmt = '%.' + '%d' % abs(digit) + 'f'

        return str(frmt % val)

    def _set_defaults(self, idict):
        """Check input params and set default values

        Look through default parameters (self.d) and given parameters (dict)
        and paste value from input if the key matches

        Parameters
        ----------
        idict : dictionary
            parameter names and values

        Modifies
        --------
            default self attributes

        """
        for key in idict:
            if hasattr(self, key):
                if key in ['cmin', 'cmax'] and type(idict[key]) != list:
                    setattr(self, key, [idict[key]])
                else:
                    setattr(self, key, idict[key])



# Name:         mapper_globcolour_l3m
# Purpose:      Mapping for L3 mapped GLOBCOLOUR data
# Authors:      Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html

from __future__ import unicode_literals, print_function, division
import datetime
import os.path
import glob

import numpy as np

from nansat.vrt import VRT
from nansat.mappers.globcolour import Globcolour
from nansat.utils import gdal, ogr
from nansat.exceptions import WrongMapperError


class Mapper(VRT, Globcolour):
    """Mapper for GLOBCOLOR L3M products"""

    def __init__(self, filename, gdalDataset, gdalMetadata, **kwargs):
        ''' GLOBCOLOR L3M VRT '''

        try:
            print("=>%s<=" % gdalMetadata['NC_GLOBAL#title'])
        except (TypeError, KeyError):
            raise WrongMapperError

        if 'GlobColour' not in gdalMetadata['NC_GLOBAL#title']:
            raise WrongMapperError

        # get list of similar (same date) files in the directory
        iDir, iFile = os.path.split(filename)
        iFileName, iFileExt = os.path.splitext(iFile)
        print('idir:', iDir, iFile, iFileName[0:30], iFileExt[0:8])

        simFilesMask = os.path.join(iDir, iFileName[0:30] + '*.nc')
        simFiles = glob.glob(simFilesMask)
        print('simFilesMask, simFiles', simFilesMask, simFiles)

        metaDict = []
        for simFile in simFiles:
            print('simFile', simFile)
            # open file, get metadata and get parameter name
            simSupDataset = gdal.Open(simFile)
            simSubDatasets = simSupDataset.GetSubDatasets()
            simWKV = None
            for simSubDataset in simSubDatasets:
                if '_mean' in simSubDataset[0]:
                    simValidSupDataset = simSupDataset
                    simGdalDataset = gdal.Open(simSubDataset[0])
                    simBand = simGdalDataset.GetRasterBand(1)
                    simBandMetadata = simBand.GetMetadata()
                    simVarname = simBandMetadata['NETCDF_VARNAME']
                    # get WKV
                    print('    simVarname', simVarname)
                    if simVarname in self.varname2wkv:
                        simWKV = self.varname2wkv[simVarname]
                        break

            # skipp adding this similar file if it is not valid
            if simWKV is None:
                continue

            metaEntry = {
                'src': {'SourceFilename': simSubDataset[0],
                        'SourceBand':  1},
                'dst': {'wkv': simWKV, 'original_name': simVarname}}

            # add wavelength and name
            longName = simBandMetadata['long_name']
            if 'Fully normalised water leaving radiance' in longName:
                simWavelength = simVarname.split('L')[1].split('_mean')[0]
                metaEntry['dst']['suffix'] = simWavelength
                metaEntry['dst']['wavelength'] = simWavelength

            # add band with rrsw
            metaEntry2 = None
            if simWKV == 'surface_upwelling_spectral_radiance_in_air_emerging_from_sea_water':
                solarIrradiance = simBandMetadata['solar_irradiance']
                metaEntry2 = {'src': metaEntry['src']}
                metaEntry2['dst'] = {'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea'
                                            '_water_to_downwelling_radiative_flux_in_water',
                                     'suffix': simWavelength,
                                     'wavelength': simWavelength,
                                     # 'expression': 'self["nLw_%s"] / %s / (0.52 + 1.7 * self["nLw
                                     # _%s"] / %s)' % (simWavelength, solarIrradiance,
                                     # simWavelength, solarIrradiance),
                                     'expression': 'self["nLw_%s"] / %s' %
                                     (simWavelength, solarIrradiance)
                                     }

            print('        metaEntry', metaEntry)
            metaDict.append(metaEntry)
            if metaEntry2 is not None:
                print('        metaEntry2', metaEntry2)
                metaDict.append(metaEntry2)

        print('simSubDatasets', simValidSupDataset.GetSubDatasets())
        for simSubDataset in simValidSupDataset.GetSubDatasets():
            print('simSubDataset', simSubDataset)
            if '_flags ' in simSubDataset[1]:
                print('    mask simSubDataset', simSubDataset[1])
                flags = gdal.Open(simSubDataset[0]).ReadAsArray()
                mask = np.ones(flags.shape) * 64
                mask[np.bitwise_and(flags, np.power(2, 0)) > 0] = 1
                mask[np.bitwise_and(flags, np.power(2, 3)) > 0] = 2

        self.band_vrts = {'maskVRT': VRT(array=mask)}

        metaDict.append(
            {'src': {'SourceFilename': self.band_vrts['maskVRT'].filename,
                     'SourceBand': 1},
             'dst': {'name': 'mask'}})

        # create empty VRT dataset with geolocation only
        simGdalDataset.SetProjection('GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,'
                                     '298.257223563,AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG",'
                                     '"6326"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],'
                                     'UNIT["degree",0.01745329251994328,AUTHORITY["EPSG","9122"]],'
                                     'AUTHORITY["EPSG","4326"]]')
        self._init_from_gdal_dataset(simGdalDataset)

        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)

        # Add valid time
        startYear = int(gdalMetadata['Start Year'])
        startDay = int(gdalMetadata['Start Day'])
        # Adding valid time to dataset
        self.dataset.SetMetadataItem('time_coverage_start',
            (datetime.datetime(startYear, 1, 1) + datetime.timedelta(startDay)).isoformat())
        self.dataset.SetMetadataItem('time_coverage_end',
            (datetime.datetime(startYear, 1, 1) + datetime.timedelta(startDay)).isoformat())



# Name:        mapper_mod44w
# Purpose:     Mapping for MOD44W watermask data
# Authors:      Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
import os.path
import json

import pythesint as pti

from nansat.vrt import VRT
from nansat.exceptions import WrongMapperError


class Mapper(VRT):
    ''' VRT with mapping of WKV for MOD44W produc (MODIS watermask at 250 m)'''

    def __init__(self, filename, gdalDataset, gdalMetadata, **kwargs):
        ''' Create VRT '''

        fileBaseName = os.path.basename(filename)
        if not fileBaseName == 'MOD44W.vrt':
            raise WrongMapperError

        metaDict = [{'src': {'SourceFilename': filename, 'SourceBand':  1},
                     'dst': {'wkv': 'land_binary_mask'}}]

        # create empty VRT dataset with geolocation only
        self._init_from_gdal_dataset(gdalDataset)

        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)

        mm = pti.get_gcmd_instrument('MODIS')
        ee = pti.get_gcmd_platform('TERRA')
        self.dataset.SetMetadataItem('instrument', json.dumps(mm))
        self.dataset.SetMetadataItem('platform', json.dumps(ee))



import sys
import importlib
import unittest
pixfun_module_name = 'nansat._pixfun_py{0}'.format(sys.version_info[0])

class TestPixelFunctions(unittest.TestCase):
    def test_import_pixel_functions(self):
        try:
            pixfun = importlib.import_module(pixfun_module_name)
            pixfun.registerPixelFunctions()
        except ImportError:
            self.fail('Cannot import pixel functions')



import json
import numpy as np
try:
    import scipy
except:
    IMPORT_SCIPY = False
else:
    IMPORT_SCIPY = True

import gdal
from netCDF4 import Dataset

import pythesint as pti

from nansat.nsr import NSR
from nansat.vrt import VRT
from nansat.utils import initial_bearing
from nansat.exceptions import WrongMapperError, NansatReadError

class Sentinel1(VRT):
    """ Mapper class to access Sentinel-1 data with netCDF4 to work for both opendap streams and
    local files.
    """
    timeVarName = 'time'
    input_filename = ''

    def __init__(self, filename, flip_gcp_line=False):
        #if not 'S1A' in filename or not 'S1B' in filename:
        #    raise WrongMapperError('%s: Not Sentinel 1A or 1B' %filename)
        if not self.dataset.GetMetadataItem('SATELLITE_IDENTIFIER') or \
                not self.dataset.GetMetadataItem('SATELLITE_IDENTIFIER').lower()=='sentinel-1':
            raise WrongMapperError('%s: Not Sentinel 1A or 1B' %filename)
        if not IMPORT_SCIPY:
            raise NansatReadError('Sentinel-1 data cannot be read because scipy is not installed')


        self.input_filename = filename
        try:
            self.ds = Dataset(filename)
        except OSError:
            self.ds = Dataset(filename+'#fillmismatch')
            self.input_filename = filename+'#fillmismatch'
        try:
            lon = self.ds.variables['GCP_longitude_'+self.ds.polarisation[:2]]
        except (AttributeError, KeyError):
            raise WrongMapperError('%s: Not Sentinel 1A or 1B' %filename)

        self._remove_geotransform()
        self._remove_geolocation()
        self.dataset.SetProjection('')
        self.dataset.SetGCPs(self.get_gcps(flip_gcp_line=flip_gcp_line), NSR().wkt)
        self.add_incidence_angle_band()
        self.add_look_direction_band()
        self.set_gcmd_dif_keywords()

    def set_gcmd_dif_keywords(self):
        mditem = 'entry_title'
        if not self.dataset.GetMetadataItem(mditem):
            self.dataset.SetMetadataItem(mditem, self.input_filename)
        mditem = 'data_center'
        if not self.dataset.GetMetadataItem(mditem):
            self.dataset.SetMetadataItem(mditem, json.dumps(pti.get_gcmd_provider('NO/MET')))
        mditem = 'ISO_topic_category'
        if not self.dataset.GetMetadataItem(mditem):
            self.dataset.SetMetadataItem(mditem,
                pti.get_iso19115_topic_category('Imagery/Base Maps/Earth Cover')['iso_topic_category'])

        mm = pti.get_gcmd_instrument('sar')
        if self.ds.MISSION_ID=='S1A':
            ee = pti.get_gcmd_platform('sentinel-1a')
        else:
            ee = pti.get_gcmd_platform('sentinel-1b')
        self.dataset.SetMetadataItem('instrument', json.dumps(mm))
        self.dataset.SetMetadataItem('platform', json.dumps(ee))

        self.dataset.SetMetadataItem('time_coverage_start',
                self.dataset.GetMetadataItem('ACQUISITION_START_TIME'))
        self.dataset.SetMetadataItem('time_coverage_end', 
        self.dataset.GetMetadataItem('ACQUISITION_STOP_TIME'))

    def get_gcps(self, flip_gcp_line=False):
        """ Get Ground Control Points for the dataset. 

        Note that OPeNDAP streams and netCDF files are read differently by gdal. The OPeNDAP streams
        are read by specifying the get parameters to the OPeNDAP url. The get parameters specify the
        reference dimensions, e.g., x and y. Since these are specified, the raster data is correctly
        referenced to the GCPs. However, when gdal reads a raster band from netCDF, it reads it
        "blindly". This is risky, since the definition of origo may be different in gdal vs the
        original data (e.g., first line starts in upper left corner or in lower left corner). For
        Sentinel-1, the raster data is flipped in relation to the GCPs, so we need to flip the GCP
        line vector as well.

        """
        lon = self.ds.variables['GCP_longitude_'+self.ds.polarisation[:2]][:]
        lat = self.ds.variables['GCP_latitude_'+self.ds.polarisation[:2]][:]
        line = self.ds.variables['GCP_line_'+self.ds.polarisation[:2]][:]
        if flip_gcp_line:
            # Flip line vector
            line = self.ds.dimensions['y'].size - line
        pixel = self.ds.variables['GCP_pixel_'+self.ds.polarisation[:2]][:]

        gcps = []
        for i0 in range(0, self.ds.dimensions['gcp_index'].size):
            gcp = gdal.GCP(float(lon[i0]), float(lat[i0]), 0, float(pixel[i0]),
                    float(line[i0]))
            gcps.append(gcp)

        return gcps

    def add_incidence_angle_band(self):
        # Get GCP variables
        pixel = self.ds['GCP_pixel_'+self.ds.polarisation[:2]][:].data
        line = self.ds['GCP_line_'+self.ds.polarisation[:2]][:].data
        inci = self.ds['GCP_incidenceAngle_'+self.ds.polarisation[:2]][:].data
        inci = inci.reshape(np.unique(line[:].data).shape[0],
                np.unique(pixel[:].data).shape[0])

        # Add incidence angle band
        inciVRT = VRT.from_array(inci)
        inciVRT = inciVRT.get_resized_vrt(self.dataset.RasterXSize, self.dataset.RasterYSize, 1)
        self.band_vrts['inciVRT'] = inciVRT
        src = {'SourceFilename': self.band_vrts['inciVRT'].filename,
               'SourceBand': 1}
        dst = {'wkv': 'angle_of_incidence',
               'name': 'incidence_angle'}
        self.create_band(src, dst)
        self.dataset.FlushCache()

    def get_full_size_GCPs(self):
        # Get GCP variables
        pixel = self.ds['GCP_pixel_' + self.ds.polarisation[:2]][:].data
        line = self.ds['GCP_line_' + self.ds.polarisation[:2]][:].data
        lon = self.ds['GCP_longitude_' + self.ds.polarisation[:2]][:].data
        lat = self.ds['GCP_latitude_' + self.ds.polarisation[:2]][:].data
        lon = lon.reshape(np.unique(line[:].data).shape[0],
                np.unique(pixel[:].data).shape[0])
        lat = lat.reshape(np.unique(line[:].data).shape[0],
                np.unique(pixel[:].data).shape[0])
        return lon, lat

    def add_look_direction_band(self):
        lon, lat = self.get_full_size_GCPs()
        """
        TODO: Also in mapper_sentinel1_l1.py... Use this code there..
        """
        sat_heading = initial_bearing(lon[:-1, :], lat[:-1, :], lon[1:, :], lat[1:, :])
        look_direction = scipy.ndimage.interpolation.zoom( np.mod(sat_heading + 90, 360),
                (np.shape(lon)[0] / (np.shape(lon)[0]-1.), 1))

        # Decompose, to avoid interpolation errors around 0 <-> 360
        look_direction_u = np.sin(np.deg2rad(look_direction))
        look_direction_v = np.cos(np.deg2rad(look_direction))
        look_u_VRT = VRT.from_array(look_direction_u)
        look_v_VRT = VRT.from_array(look_direction_v)
        lookVRT = VRT.from_lonlat(lon, lat)
        lookVRT.create_band([{'SourceFilename': look_u_VRT.filename,
                               'SourceBand': 1},
                              {'SourceFilename': look_v_VRT.filename,
                               'SourceBand': 1}],
                             {'PixelFunctionType': 'UVToDirectionTo'}
                             )

        # Blow up to full size
        lookVRT = lookVRT.get_resized_vrt(self.dataset.RasterXSize, self.dataset.RasterYSize, 1)

        # Store VRTs so that they are accessible later
        self.band_vrts['look_u_VRT'] = look_u_VRT
        self.band_vrts['look_v_VRT'] = look_v_VRT
        self.band_vrts['lookVRT'] = lookVRT

        src = {
                'SourceFilename': self.band_vrts['lookVRT'].filename,
                'SourceBand': 1
            }
        dst = {
                'wkv': 'sensor_azimuth_angle',
                'name': 'look_direction'
            }
        self.create_band(src, dst)
        self.dataset.FlushCache()
        """ End repetition """



# Name:        mapper_kmssL1
# Purpose:     Mapping for KMSS-L1 data
# Author:      Evgeny Morozov
# Licence:     This file is part of NANSAT. You can redistribute it or modify
#              under the terms of GNU General Public License, v.3
#              http://www.gnu.org/licenses/gpl-3.0.html
import os
from datetime import datetime

from nansat.utils import gdal, ogr
from nansat.exceptions import WrongMapperError
from nansat.vrt import VRT
from nansat.domain import Domain


class Mapper(VRT):
    ''' VRT with mapping of WKV for KMSS TOA tiff data'''

    def __init__(self, filename, gdalDataset, gdalMetadata, **kwargs):
        ''' Create VRT '''
        if (os.path.split(filename)[1][0:4] != '101_' or
                os.path.split(filename)[1][0:4] != '102_'):
                raise WrongMapperError

        try:
            product = gdalDataset.GetDriver().LongName
        except:
            raise WrongMapperError

        if (product != 'GeoTIFF' or filename[-3:] != 'tif' or
                gdalDataset.RasterCount != 3):
            raise WrongMapperError

        metaDict = [{'src': {'SourceFilename': filename, 'SourceBand': 1},
                     'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                             'wavelength': '555'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 2},
                     'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                             'wavelength': '655'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 3},
                     'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                             'wavelength': '800'}}
                    ]
        # from https://gsics.nesdis.noaa.gov/wiki/Development/StandardVariableNames

        # add DataType into 'src' and name into 'dst'
        for bandDict in metaDict:
            if 'DataType' not in bandDict['src']:
                bandDict['src']['DataType'] = 2
            if 'wavelength' in bandDict['dst']:
                bandDict['dst']['name'] = ('toa_radiance_' +
                                           bandDict['dst']['wavelength'])

        # create empty VRT dataset with geolocation only
        self._init_from_gdal_dataset(gdalDataset)

         # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)
